/* cluster: routines for dealing with clusters of vectors
 *
 * Author: Dylan Muir (dr.muir@student.qut.edu.au)
 *         QUT MLRC LPG Semester 2 1999
 * Date: 14th September, 1999
 * Modified: 21st November, 2000
 * Version: 0.05
 */

#ifndef	__cluster_h
#define	__cluster_h

/* -- required modules -- */

#include <stdio.h>
#include "vector_utils.h"
#include "stddefs.h"

/* -- Cluster structure -- */

struct clusterTag {
	vector				*vect;
	struct clusterTag	*next;
};

typedef struct clusterTag cluster;


/* -- Cluster functions -- */

/* -- Function ClusterNew
 * Pre: TRUE
 * Post: (a new empty cluster was returned)
 * Note: a new cluster is always empty
 */
cluster	*ClusterNew(void);


/* -- Function ClusterDestroy
 * Pre: 'clust' is a valid allocated cluster
 * Post: 'clust' was destroyed && a NULL cluster was returned
 */
cluster	*ClusterDestroy(cluster *clust);


/* -- Function ClusterAddVector
 * Pre: 'clust' is a valid allocated cluster
 *      'vect' is a valid allocated vector
 * Post: a cluster containing both 'clust' and 'vect' was returned
 */
cluster	*ClusterAddVector(cluster *clust, vector *vect);


/* -- Function ClusterRemoveVector
 * Pre: 'clust' is a valid allocated cluster
 *      'vect' is a valid allocated vector
 * Post: a cluster with 'vect' removed was returned
 * Note: if two instances of 'vect' exist in 'clust', only one will be removed
 */
cluster *ClusterRemoveVector(cluster *clust, vector *vect);


/* -- Function ClusterMerge
 * Pre: 'clust1' and 'clust2' are valid allocated clusters
 * Post: a cluster containing both 'clust1' and 'clust2' was returned
 */
cluster *ClusterMerge(cluster *clust1, cluster *clust2);


/* -- Function ClusterCentroid
 * Pre: 'clust' is a valid allocated non-empty cluster
 * Post: the centroid of 'clust' was calcluated and returned
 * Note: this function allocates a new vector, which you
 *       MUST destroy
 */
vector	*ClusterCentroid(cluster *clust);


/* -- Function ClusterSize
 * Pre: 'clust' is a valid allocated cluster
 * Post: the number of vectors in 'clust' was returned
 */
int	ClusterSize(cluster *clust);


/* -- Function ClusterIsEmpty
 * Pre: 'clust' is a valid allocated cluster
 * Post: returned (ClusterSize(clust) == 0)
 */
bool	ClusterIsEmpty(cluster *clust);


/* -- Function ClusterReadFromFile
 * Pre: 'inputFile' is a vector file opened for reading
 * Post: All the vectors from 'inputFile' have been read and added to a cluster
 *       which is subsequently returned.  If no vectors could be read, an
 *       empty cluster is returned.
 * Note: This relies on VectorReadLine.  See vector_read.h for details
 */
cluster *ClusterReadFromFile(FILE *inputFile);


/* -- Function ClusterCheckDimensionality
 * Pre: 'clust' is a valid allocated cluster
 * Post: (all vectors in 'clust' have the same dimensions && TRUE was returned) ||
 *       (at least one vector had a different number of dimensions && FALSE was returned)
 */
bool ClusterCheckDimensionality(cluster *clust);


/* -- Function ClusterVectorIsIn
 * Pre: 'clust' is a valid allocated cluster
 *      'vect' is a valid allocated vector
 * Post: ('vect' is in 'clust' && TRUE was returned) ||
 *       ('vect' was not found in 'clust' && FALSE was returned)
 */
bool ClusterVectorIsIn(cluster *clust, vector *vect);


/* -- Function ClusterMakeCopy
 * Pre: 'clust' is the cluster to copy
 * Post: a copy of 'clust' was made and returned
 */
cluster *ClusterMakeCopy(cluster *clust);

#endif	/* __cluster_h */

/* --- END of cluster.h --- */


