/* Vector utils: a series of functions for dealing with vectors
 *
 * SEE vector_utils.h for details
 */
#include "vector_utils.h"
#include "2darray.h"
#include <stdlib.h>
#include <math.h>
#include <memory.h>
#ifdef	UNIX
	#include <ieeefp.h>
#endif
#ifdef	WIN32
	#include	<float.h>
	#define	FP_SNAN	_FPCLASS_SNAN
#endif

#include <stdio.h>

/* -- Vector macros -- */

#define	SQR(x)	((x) * (x))
#define	ABS(x)	((x) < 0 ? (-(x)) : (x))


/* -- Vector functions -- */

vector *VectorAllocate(int dimensions)
{
	vector	*vect;

	if ((vect = (vector *) malloc(sizeof(vector))) == NULL)
		return NULL;

	if ((vect -> components = (double *) malloc(dimensions * sizeof(double))) == NULL) {
		free(vect);
		return NULL;
	}

	vect -> dimensions = dimensions;
	return vect;
}

vector *VectorConstruct(int dimensions, const double *components)
{
	vector	*newVect;

	if (!(newVect = VectorAllocate(dimensions)))
		return NULL;

	memcpy(newVect -> components, components, dimensions * sizeof(double));
	return newVect;
}


vector *VectorDeallocate(vector *vect)
{
	free(vect -> components);
	free(vect);
	return NULL;
}

double VectorLength(vector *vect)
{
	vector	*zero;
	double	length;

	if ((zero = VectorZero(vect -> dimensions)) == NULL)
		return 0.0;

	length = VectorEuclideanDist(vect, zero);
	VectorDeallocate(zero);
	return length;
}

double VectorEuclideanDist(vector *vect1, vector *vect2)
{
	double	dimensionSum;
	int	dimension;

	if (vect1 -> dimensions != vect2 -> dimensions)
		return 0.0;

	dimensionSum = 0.0;
	dimension = 0;
	while (dimension < vect1 -> dimensions) {
		dimensionSum += SQR(vect1 -> components[dimension] - vect2 -> components[dimension]);
		dimension++;
	}

	return sqrt(dimensionSum);
}

vector *VectorZero(int dimensions)
{
	vector	*zero;
	int	compIndex;		/* index into components */

	if ((zero = VectorAllocate(dimensions)) == NULL)
		return NULL;

	compIndex = 0;
	while (compIndex < dimensions) {
		zero -> components[compIndex] = 0.0;
		compIndex++;
	}
	return zero;
}

vector *VectorUnit(int dimensions)
{
	vector	*unit;
	int	compIndex;		/* index into componenets */

	if ((unit = VectorAllocate(dimensions)) == NULL)
		return NULL;

	compIndex = 0;
	while (compIndex < dimensions) {
		unit -> components[compIndex] = 1 / (double) dimensions;
		compIndex++;
	}
	return unit;
}

double VectorDotProduct(vector *vect1, vector *vect2)
{
	double	productSum;
	int	dimension;

	if (vect1 -> dimensions != vect2 -> dimensions)
		return 0.0;

	productSum = 0.0;
	dimension = 0;
	while (dimension < vect1 -> dimensions) {
		productSum += vect1 -> components[dimension] * vect2 -> components[dimension];
		dimension++;
	}

	return productSum;
}

void VectorSum(vector *vectd, vector *vects)
{
	int	dIndex;

	if (vectd -> dimensions != vects -> dimensions)
		return;

	dIndex = vectd -> dimensions - 1;
	while (dIndex >= 0) {
		vectd -> components[dIndex] += vects -> components[dIndex];
		dIndex--;
	}
}

void VectorSub(vector *vectd, vector *vects)
{
	int	dIndex;

	if (vectd -> dimensions != vects -> dimensions)
		return;

	dIndex = vectd -> dimensions - 1;
	while (dIndex >= 0) {
		vectd -> components[dIndex] -= vects -> components[dIndex];
		dIndex--;
	}
}

void VectorNeg(vector *vect)
{
	int	dIndex;

	dIndex = vect -> dimensions - 1;
	while (dIndex >= 0) {
		vect -> components[dIndex] = -(vect -> components[dIndex]);
		dIndex--;
	}
}

void VectorSquare(vector *vect)
{
	int	dIndex;

	dIndex = vect -> dimensions - 1;
	while (dIndex >= 0) {
		vect -> components[dIndex] *= vect -> components[dIndex];
		dIndex--;
	}
}


void VectorMultiplyScalar(vector *vect, double scalar)
{
	int	dIndex,
		dimensions;			/* for speedup */

	dimensions = vect -> dimensions;
	dIndex = 0;
	while (dIndex < dimensions) {
		vect -> components[dIndex] *= scalar;
		dIndex++;
	}
}

void VectorDivideScalar(vector *vect, double scalar)
{
	VectorMultiplyScalar(vect, 1.0 / scalar);
}


vector *VectorMakeCopy(vector *vect)
{
	vector	*newVect;
	int	dIndex;		/* index into components (across dimensions) */

	if (vect == NULL)
		return NULL;

	if (!(newVect = VectorAllocate(vect -> dimensions)))
		return NULL;

	dIndex = 0;
	while (dIndex < vect -> dimensions) {
		newVect -> components[dIndex] = vect -> components[dIndex];
		dIndex++;
	}

	return newVect;
}


vector *VectorMakeHomogeneous(vector *vect)
{
	vector	*homVect;
	int		dIndex;		/* index into components (across dimensions) */

	if (vect == NULL)
		return NULL;

	if (!(homVect = VectorAllocate(vect -> dimensions + 1)))
		return NULL;

	dIndex = 0;
	while (dIndex < vect -> dimensions) {
		homVect -> components[dIndex] = vect -> components[dIndex];
		dIndex++;
	}

	homVect -> components[vect -> dimensions] = 1.0;

	return homVect;
}


int VectorMaxComponent(vector *vect)
{
	int	indexMax,
		cIndex;
	double	maxComponent;

	maxComponent = vect -> components[0];
	indexMax = 0;
	cIndex = 1;
	while (cIndex < vect -> dimensions) {
		if (vect -> components[cIndex] > maxComponent) {
			maxComponent = vect -> components[cIndex];
			indexMax = cIndex;
		}
		cIndex++;
	}

	return indexMax;
}

bool VectorIdentical(vector *vect1, vector *vect2)
{
	int	cIndex;

	if (vect1 -> dimensions != vect2 -> dimensions)
		return FALSE;

	cIndex = 0;
	while (cIndex < vect1 -> dimensions) {
		if (vect1 -> components[cIndex] != vect2 -> components[cIndex])
			return FALSE;
		cIndex++;
	}

	return TRUE;
}

double VectorManhattanEntropy(vector *vect1, vector *vect2)
{
/*
 *    H = -Sum(d)[fd * logn(fd)]
 *
 *              |ad - bd|
 *    fd = -------------------
 *          Sum(d)[|ad - bd|]
 *
 */
	double	mDistSum, fd, hSum;
	int	dIndex;

	if (vect1 -> dimensions != vect2 -> dimensions)
		return FP_SNAN;

	mDistSum = 0.0;
	for (dIndex = 0; dIndex < vect1 -> dimensions; dIndex++)
		mDistSum += ABS(vect1 -> components[dIndex] - vect2 -> components[dIndex]);

	if (mDistSum == 0.0)
		return 0.0;

	hSum = 0.0;
	for (dIndex = 0; dIndex < vect1 -> dimensions; dIndex++) {
		fd = ABS(vect1 -> components[dIndex] - vect2 -> components[dIndex]) / mDistSum;
		if (fd != 0.0) hSum -= fd * log(fd);
	}

	return hSum;
}


double **VectorCorrelation(vector *vect, int *matrixSize)
{
	double	**matrix;
	int	xIndex, yIndex;
	double	component;

	*matrixSize = vect -> dimensions;
	if (!Allocate2DArray((char ***) &matrix, *matrixSize, *matrixSize, sizeof(double)))
		return NULL;

	for (yIndex = 0; yIndex < *matrixSize; yIndex++) {
		component = vect -> components[yIndex];
		for (xIndex = 0; xIndex < *matrixSize; xIndex++)
			matrix[xIndex][yIndex] = component * vect -> components[xIndex];
	}

	return matrix;
}

/* --- END of vector_utils.c --- */


